unit UJoinTables;

interface
uses __FlexCelTypes, FlexCel.Core, FlexCel.Report, _UVirtualDataTable, SysUtils, Generics. Collections;
type
  TJoinTables = class(TVirtualDataTable)
  private
    Tables: TObjectList<TVirtualDataTable>;
  protected
    function Get_Locale: CultureInfo; override;
    function Get_ColumnCount: Int32; override;

  public
    constructor Create(const aName: string; const aTables: array of TVirtualDataTable);
    destructor Destroy; override;

    function CreateState(const sort: UTF16String; const masterDetailLinks: TMasterDetailLinkArray; const splitLink: TSplitLink): TVirtualDataTableState; override;
    function GetColumn(const columnName: UTF16String): Int32; override;
    function GetColumnName(const columnIndex: Int32): UTF16String; override;
    function GetColumnCaption(const columnIndex: Int32): UTF16String; override;

  end;

  TJoinTablesState = class(TVirtualDataTableState)
  private
    TableStates: TObjectList<TVirtualDataTableState>;
    ColsPerTable: TArray<integer>;
    CachedRowCount: integer;
    function FindTableState(const column: integer;
      out offset: integer): TVirtualDataTableState;
  protected
    function Get_RowCount: Int32; override;
  public
    constructor Create(const aTables: TObjectList<TVirtualDataTable>; const sort: UTF16String; const masterDetailLinks: TMasterDetailLinkArray; const splitLink: TSplitLink);
    destructor Destroy; override;
    function GetValue(const column: Int32): TReportValue; override;
    function GetValue(const row: Int32; const column: Int32): TReportValue; override;
    procedure MoveFirst; override;
    procedure MoveNext; override;
  end;



implementation

{ TJoinTables<T, K> }
constructor TJoinTables.Create(const aName: string; const aTables: array of TVirtualDataTable);
var
  i: Integer;
begin
  inherited Create(aName, nil);
  Tables := TObjectList<TVirtualDataTable>.Create;
  Tables.AddRange(aTables);

end;

function TJoinTables.CreateState(const sort: UTF16String;
  const masterDetailLinks: TMasterDetailLinkArray;
  const splitLink: TSplitLink): TVirtualDataTableState;
begin
  Result := TJoinTablesState.Create(Tables, sort, masterDetailLinks, splitLink);
end;

destructor TJoinTables.Destroy;
begin
  Tables.Free;
  inherited;
end;

function TJoinTables.GetColumn(const columnName: UTF16String): Int32;
var
  i, k, acum, SepPos: integer;
  prefix: string;
begin
  acum := 0;
  SepPos := Pos('_', columnName);
  if SepPos < 1 then exit(-1);

  for i := 0 to Tables.Count - 1 do
  begin
    prefix := System.Copy(columnName, 1, SepPos - 1);
    if prefix = Tables[i].TableName then
    begin
      k := Tables[i].GetColumn(Copy(columnName, SepPos + 1, Length(ColumnName)));
      if (k >= 0) then exit(acum + k);
    end;
    Inc(Acum, Tables[i].ColumnCount);
  end;

  Result := -1;
end;

function TJoinTables.GetColumnCaption(
  const columnIndex: Int32): UTF16String;
var
  i: integer;
  cc: integer;
begin
  cc := 0;
  for i := 0 to Tables.Count - 1 do
  begin
    if columnIndex - cc < Tables[i].ColumnCount then exit(Tables[i].GetColumnCaption(columnIndex - cc));
    Inc(cc, Tables[i].ColumnCount);
  end;
  raise Exception.Create('Invalid column caption');
end;

function TJoinTables.GetColumnName(const columnIndex: Int32): UTF16String;
var
  i: integer;
  cc: integer;
begin
  cc := 0;
  for i := 0 to Tables.Count - 1 do
  begin
    if columnIndex - cc < Tables[i].ColumnCount then exit(Tables[i].GetColumnName(columnIndex - cc));
    Inc(cc, Tables[i].ColumnCount);
  end;
  raise Exception.Create('Invalid column index');
end;

function TJoinTables.Get_ColumnCount: Int32;
var
  i: integer;
begin
  Result := 0;
  for i := 0 to Tables.Count - 1 do
  begin
    Inc(Result, Tables[i].ColumnCount);
  end;
end;

function TJoinTables.Get_Locale: CultureInfo;
begin
  Result := CultureInfo.CurrentCulture;
end;

{ TJoinTablesState }

constructor TJoinTablesState.Create(const aTables: TObjectList<TVirtualDataTable>; const sort: UTF16String; const masterDetailLinks: TMasterDetailLinkArray; const splitLink: TSplitLink);
var
  i: integer;
begin
   TableStates := TObjectList<TVirtualDataTableState>.Create;
   SetLength(ColsPerTable, aTables.Count);
   CachedRowCount := 0;
   for i := 0 to aTables.Count - 1 do
   begin
     TableStates.Add(aTables[i].CreateState(sort, masterDetailLinks, splitLink));
     if TableStates[i].RowCount > CachedRowCount then CachedRowCount := TableStates[i].RowCount;
     ColsPerTable[i] := aTables[i].ColumnCount;
   end;
end;

destructor TJoinTablesState.Destroy;
begin
  TableStates.Free;
  inherited;
end;

function TJoinTablesState.FindTableState(const column: integer; out offset: integer): TVirtualDataTableState;
var
  i: integer;
begin
  offset := 0;
   for i := 0 to TableStates.Count - 1 do
   begin
     if column - offset < ColsPerTable[i]  then exit(TableStates[i]);
     inc(offset, ColsPerTable[i]);
   end;

   Result := nil;

end;

function TJoinTablesState.GetValue(const row, column: Int32): TReportValue;
var
  i, offset: integer;
  v: TVirtualDataTableState;
begin
  v := FindTableState(column, offset);
  if v = nil then exit(TReportValue.Empty);
  if v.Eof then exit(TReportValue.Empty);

  Result := v.GetValue(row, column - offset);

end;

function TJoinTablesState.GetValue(const column: Int32): TReportValue;
var
  i, offset: integer;
  v: TVirtualDataTableState;
begin
  v := FindTableState(column, offset);
  if v = nil then exit(TReportValue.Empty);
  if v.Eof then exit(TReportValue.Empty);

  Result := v.GetValue(column - offset);
end;

function TJoinTablesState.Get_RowCount: Int32;
begin
  Result := CachedRowCount;
end;

procedure TJoinTablesState.MoveFirst;
var
  i: integer;
begin
   for i := 0 to TableStates.Count - 1 do
   begin
     TableStates[i].MoveFirst;
   end;
end;

procedure TJoinTablesState.MoveNext;
var
  i: integer;
begin
   for i := 0 to TableStates.Count - 1 do
   begin
     if (not TableStates[i].Eof) then TableStates[i].DoMoveNext;
   end;
end;

end.
